--------------------------------------
-- 日本語文字列を出力するモジュール --
-- Programed by Ken Hashimoto       --
-- Version: 071020                  --
--------------------------------------

function han2zen(u)
  -- 半角文字を表す Unicode u に対応する全角文字の
  -- Unicode を返す。変換失敗なら nil を返す
  if u == 32 then return 12288
  elseif u == 34 then return 8243
  elseif u == 39 then return 8242
  elseif u == 92 then return 65509
  elseif u >= 33 and u <= 126 then return u + 65248
  elseif u == 65377 then return 12290
  elseif u == 65378 then return 12300
  elseif u == 65379 then return 12301
  elseif u == 65380 then return 12289
  elseif u == 65381 then return 12539
  elseif u == 65382 then return 12530
  elseif u == 65383 then return 12449
  elseif u == 65384 then return 12451
  elseif u == 65385 then return 12453
  elseif u == 65386 then return 12455
  elseif u == 65387 then return 12457
  elseif u == 65388 then return 12515
  elseif u == 65389 then return 12517
  elseif u == 65390 then return 12519
  elseif u == 65391 then return 12483
  elseif u == 65392 then return 12540
  elseif u == 65393 then return 12450
  elseif u == 65394 then return 12452
  elseif u == 65395 then return 12454
  elseif u == 65396 then return 12456
  elseif u == 65397 then return 12458
  elseif u == 65398 then return 12459
  elseif u == 65399 then return 12461
  elseif u == 65400 then return 12463
  elseif u == 65401 then return 12465
  elseif u == 65402 then return 12467
  elseif u == 65403 then return 12469
  elseif u == 65404 then return 12471
  elseif u == 65405 then return 12473
  elseif u == 65406 then return 12475
  elseif u == 65407 then return 12477
  elseif u == 65408 then return 12479
  elseif u == 65409 then return 12481
  elseif u == 65410 then return 12484
  elseif u == 65411 then return 12486
  elseif u == 65412 then return 12488
  elseif u >= 65413 and u <= 65418 then return u - 52923
  elseif u == 65419 then return 12498
  elseif u == 65420 then return 12501
  elseif u == 65421 then return 12504
  elseif u == 65422 then return 12507
  elseif u >= 65423 and u <= 65427 then return u - 52913
  elseif u == 65428 then return 12516
  elseif u == 65429 then return 12518
  elseif u >= 65430 and u <= 65435 then return u - 52910
  elseif u == 65436 then return 12527
  elseif u == 65437 then return 12531
  elseif u == 65438 then return 12443
  elseif u == 65439 then return 12444
  elseif u == 28 then return 8594
  elseif u == 29 then return 8592
  elseif u == 30 then return 8593
  elseif u == 31 then return 8595
  elseif u == 127 then return 65340
  else return nil
  end
end

function unicode2utf8(u)
  -- Unicode値をUTF-8の文字に変換
  local u0, u2, rv
  if u < 128 then return string.char(u) end
  u0 = 128
  u2 = 64
  rv = ''
  while u >= u2 do
    local u1 = math.mod(u, 64)
    rv = string.char(128 + u1) .. rv
    u = math.floor(u / 64)
    u0 = u0 / 2 + 128
    u2 = u2 / 2
  end
  return string.char(u0 + u) .. rv
end

function box(BG8Bit, x0, y0, x1, y1, c)
  -- Line Box Fill
  if y0 > y1 then do
    local y
    y = y0
    y0 = y1
    y1 = y
    end
  end
  for y = y0, y1 do
    BG8Bit:Line(x0, y, x1, y, c)
  end
end

function utf8top(str)
  -- 引数
  --   str: UTF-8文字列
  -- 返り値
  --   第1: 先頭文字(UTF-8)
  --   第2: 残りの(2文字目以降のUTF-8)文字列
  local c = string.byte(str)
  local rv = string.char(c)
  str = string.sub(str, 2)
  if c < 128 then return rv, str end -- 1バイト
  while str ~= '' do
    -- 複数バイト
    c = string.byte(str)
    if c < 128 or math.floor(c / 64) ~= 2 then break end
    rv = rv .. string.char(c)
    str = string.sub(str, 2)
  end
  return rv, str
end

function utf82unicode(str)
  -- 引数
  --   str: UTF-8文字列
  -- 返り値
  --   第1: 先頭文字のUnicode
  --   第2: 残りの(2文字目以降のUTF-8)文字列
  local ch
  local bold = 0
  local rv = 0
  ch, str = utf8top(str)
  while ch ~= '' do
    local c = string.byte(ch)
    local b0 = 128
    local b1 = 0
    ch = string.sub(ch, 2)
    for b = 7, 0, -1 do
      local c0 = math.floor(c / b0)
      c = math.mod(c, b0)
      if c0 == 0 then do
        b1 = b0
        break
        end
      end
      b0 = b0 / 2
    end
    rv = rv * b1 + c
    bold = b1
  end
  return rv, str
end

function utf8chop(str)
  -- UTF-8文字列の最後の1文字を削除する
  while str ~= '' do
    local u
    u = string.byte(str, -1)
    str = string.sub(str, 1, -2)
    if math.floor(u / 64) ~= 2 then break end
  end
  return str
end

function utf8last(str)
  -- UTF-8文字列の最後の1文字を返す
  local r = ''
  if str == '' then return r end
  while str ~= '' do
    local u
    u = string.byte(str, -1)
    r = string.char(u) .. r
    str = string.sub(str, 1, -2)
    if math.floor(u / 64) ~= 2 then break end
  end
  return r
end

function utf8len(str)
  -- UTF-8文字列の長さ(半角=1,全角=2とする)を求める
  local le = 0
  while str ~= '' do
    local u
    u, str = utf82unicode(str)
    if u < 128 or (65376 <= u and u < 65440) then c = 1  -- 半角
    else c = 2                                           -- 全角
    end
    le = le + c
  end
  return le
end

function symbol_sub(fp, wd, BG8Bit, x0, xswd, y0, cl)
  -- symbol()の下請け関数
  local c
  for i = 0, 11 do
    for j = 0, 8, 8 do
      c = fp:read(1)
      c = string.byte(c)
      for b = 7, 0, -1 do
        local d
        d = math.mod(c, 2)
        c = math.floor(c / 2)
        if d ~= 0 then do
          if j + b < wd then do
            BG8Bit:Plot(x0 + math.floor((j + b) * xswd), y0 + i, cl)
            end
          end
          end
        end
      end
    end
  end
end

function symbol(BG8Bit, x0, y0, cl, str)
  -- 日本語文字列を表示する
  --  x0, y0 : グラフィック座標
  --  cl     : 色
  --  str    : UTF-8文字列
  --  返り値 : 最後に表示した文字の直後の座標(X,Y)
  local fp
  fp = io.open('/cf/scfont.fon', 'rb')
  if fp == nil then fp = io.open('scfont.fon', 'rb') end
  while str ~= '' do
    local u, c
    local xs       -- 表示横ドット数
    c, str = utf8top(str)
    xs = utf8len(c) * 6
    u = utf82unicode(c)
    if x0 + xs > 256 then do
      x0 = 0
      y0 = y0 + 12
      end
    end
    if y0 > 180 then do
      x0 = 0
      y0 = 0
      end
    end
    if fp ~= nil then do
      local w
      local wd     -- フォントデータ上の横ドット数
      if 65376 <= u and u < 65440 then w = 0
      else do
        wd = xs
        fp:seek('set', u * 2)
        c = fp:read(1)
        w = string.byte(c)
        c = fp:read(1)
        w = w + string.byte(c) * 256
        end
      end
      if w == 0 and xs == 6 then do
        u = han2zen(u)
        if u then do
          wd = 12
          fp:seek('set', u * 2)
          c = fp:read(1)
          w = string.byte(c)
          c = fp:read(1)
          w = w + string.byte(c) * 256
          end
        else do
          box(BG8Bit, x0, y0, x0 + 5, y0 + 11, cl)
          wd = 0
          end
        end
        end
      end
      if wd > 0 then do
        local xswd = xs / wd
        w = w * 24
        w = w + 131072
        fp:seek('set', w)
        symbol_sub(fp, wd, BG8Bit, x0, xswd, y0, cl)
        end
      end
      end
    else do -- フォントファイルのオープンに失敗している場合
      box(BG8Bit, x0, y0, x0 + xs - 1, y0 + 11, cl)
      end
    end
    x0 = x0 + xs
  end
  if fp ~= nil then fp:close() end
  return x0, y0
end
